/* zgsmplay - a real-time curses interface for playing gsm files.
 * Copyright (C) 1995-1999 Russell Marks. See zgsmplay.c for license details.
 *
 * filesel.c - the file selector.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <curses.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "zgsmplay.h"
#include "readnbkey.h"
#include "filesel.h"


#define MAX_FLEN	20
#define MAX_DLEN	(COLS-4)
#define BARWIDTH	(MAX_FLEN+5)
#define XSIZ		((COLS-2)/BARWIDTH)
#define YSIZ		(LINES-4)

#define fwinxpos(f)	(2+BARWIDTH*((f-1)/YSIZ))
#define fwinypos(f)	(3+((f)-1)%YSIZ)

struct gsmdir_tag {
  char name[256];
  char isdir;            /* 0=file, 1=dir. */
  char marked;
  };
int gsmdirsiz;

struct gsmdir_tag *gsmdir=NULL;
int gsmdir_byte_size=64*sizeof(struct gsmdir_tag);
int gsmdir_byte_incr=32*sizeof(struct gsmdir_tag);


int tagview_mode=0;
int nodelprompt=0;




/* allocate some initial space for the gsmdir[] array.
 * This should be run before starting up curses.
 * (It's called by main(), in fact.)
 */
void gsmdir_init()
{
if(gsmdir!=NULL) return;	/* sanity check */

if((gsmdir=malloc(gsmdir_byte_size))==NULL)
  fprintf(stderr,"zgsmplay: out of memory\n"),exit(1);
}


/* make gsmdir bigger if needed.
 * call this *before writing each new entry*.
 * if this returns zero, the caller should quit with not-enough-memory
 * error, exiting curses with screenoff() if needed.
 *
 * note that gsmdir is never made smaller. This would complicate things
 * considerably for a fairly small gain, I think.
 */
int gsmdir_resize_if_needed(int newent)
{
/* this is absurdly conservative, just to be on the safe side :-) */
if((newent+1)*sizeof(struct gsmdir_tag)>=gsmdir_byte_size)
  {
  gsmdir_byte_size+=gsmdir_byte_incr;
  if((gsmdir=realloc(gsmdir,gsmdir_byte_size))==NULL)
    return(0);		/* out of memory */
  }

return(1);		/* worked */
}


void fselheadfoot()
{
clearscreen();

/* draw screen frame */
if(line_chars)
  draw_box(0,0,COLS,LINES,
    ACS_HLINE,ACS_VLINE,
    ACS_ULCORNER,ACS_URCORNER,
    ACS_LLCORNER,ACS_LRCORNER);
else
  draw_box(0,0,COLS,LINES, '-','|', ',','.', '`','\'');

stashcursor();
refresh();
}


void printmax(char *str,int maxlen)
{
static char buf[256];

strcpy(buf,str);
buf[maxlen]=0;
addstr(buf);
}


void printmaxspc(char *str,int maxlen)
{
static char buf[256];
int f;

for(f=0;f<strlen(str);f++)
  buf[f]=32;
buf[f]=0;

buf[maxlen]=0;
addstr(buf);
}


/* this could also be file not found of course */
int permissiondenied(char *fname)
{
FILE *junk;
if((junk=fopen(fname,"rb"))==NULL)
  return(1);
fclose(junk);
return(0);
}


void prettyfile(char *buf,struct gsmdir_tag *gsmdptr)
{
if(gsmdptr->isdir)
  {
  buf[0]='(';
  strcpy(buf+1,gsmdptr->name);
  strcat(buf,")");
  }
else
  strcpy(buf,gsmdptr->name);
}


void showbar(int entnum,int selected,int startfrom)
{
char ctmp[100];
int xpos,ypos;

xpos=fwinxpos(entnum-startfrom+1);
if((xpos<1)||(xpos+BARWIDTH>COLS)) return;
ypos=fwinypos(entnum-startfrom+1);
prettyfile(ctmp,&(gsmdir[entnum]));

move(ypos,xpos);
if(gsmdir[entnum].marked) addch('*'); else addch(' ');
if(selected) addch('['); else addch(' ');
printmax(gsmdir[entnum].name,MAX_FLEN);
if(selected) addch(']'); else addch(' ');
refresh();
}



int load_single_file(int curent)
{
filename=gsmdir[curent].name;

#if 0
if(!gsmplay())
  {
  /* XXX report error */
  return(0);
  }
#else
gsmplay();
#endif

return(1);
}


void load_tagged_files()
{
int f,t;

for(f=1,t=0;f<=gsmdirsiz;f++)
  if(gsmdir[f].marked) t++;

if(t==0)
  {
  return;
  }

/* zgsmplay.c sets tagview_mode==0 if esc is pressed */
tagview_mode=1;
do
  {
  for(f=1;f<=gsmdirsiz && tagview_mode;f++)
    {
    if(gsmdir[f].marked==0) continue;
    if(permissiondenied(gsmdir[f].name))
      {
      if(f>1) fselheadfoot();
      return;
      }
    if(load_single_file(f)==0) break;
    }
  }
while (autorepeat && sgres!=1);
}


void showgsmdir(int startfrom,int unshow)
{
static char cdir[MAXPATHLEN+1];
static char ctmp[1024];
int f,ypos,xpos;

getcwd(cdir,MAXPATHLEN);
sprintf(ctmp,"Directory of %s",cdir);
move(1,2);
if(unshow)
  printmaxspc(ctmp,MAX_DLEN);
else
  printmax(ctmp,MAX_DLEN);

for(f=startfrom;f<=gsmdirsiz;f++)
  {
  xpos=fwinxpos(f-startfrom+1);
  if(xpos+BARWIDTH>COLS) break;
  ypos=fwinypos(f-startfrom+1);
  prettyfile(ctmp,&(gsmdir[f]));

  move(ypos,xpos);
  if(gsmdir[f].marked && !unshow) addch('*'); else addch(' ');
  addch(' ');
  if(unshow)
    printmaxspc(gsmdir[f].name,MAX_FLEN);
  else
    printmax(gsmdir[f].name,MAX_FLEN);
  addch(' ');
  }
refresh();
}


int gcompare(void *gn1,void *gn2)
{
struct gsmdir_tag *g1,*g2;
char buf1[80],buf2[80];

g1=(struct gsmdir_tag *)gn1; g2=(struct gsmdir_tag *)gn2;
prettyfile(buf1,g1); prettyfile(buf2,g2);
return(strcmp(buf1,buf2));
}


void readgsmdir()
{
static char cdir[MAXPATHLEN+1];
DIR *dirfile;
struct dirent *anentry;
struct stat buf;
int entnum,l,isdir;

getcwd(cdir,MAXPATHLEN);
dirfile=opendir(".");
entnum=0;
while((anentry=readdir(dirfile))!=NULL)
  {
  if(anentry->d_name[0]=='.' && anentry->d_name[1]!='.')
    continue;	/* skip (most) 'dot' files */
  if((strcmp(anentry->d_name,".")!=0) &&	/* no . and no .. if at root */
     (!((strcmp(cdir,"/")==0)&&(strcmp(anentry->d_name,"..")==0))))
    {
    if((stat(anentry->d_name,&buf))==-1)
      buf.st_mode=0;
    isdir=S_ISDIR(buf.st_mode);
    /* directories, gsm tested here */
    if(((l=strlen(anentry->d_name))>4) || isdir)
      if((!strcasecmp(anentry->d_name+l-4,".gsm")) || isdir)
        {
        entnum++;
        if(!gsmdir_resize_if_needed(entnum))
          {
          screenoff();
          fprintf(stderr,"zgsmplay: out of memory\n");
          exit(1);
          }
        
        strcpy(gsmdir[entnum].name,anentry->d_name);
        gsmdir[entnum].isdir=isdir;
        gsmdir[entnum].marked=0;
        }
    }
  }
closedir(dirfile);
gsmdirsiz=entnum;
qsort(&(gsmdir[1]),gsmdirsiz,sizeof(struct gsmdir_tag),(void *)gcompare);
}


void redrawall(int curent,int startfrom)
{
fselheadfoot();
showgsmdir(startfrom,0);
showbar(curent,1,startfrom);
}


int delete_file(char *filename)
{
/* XXX NYI */
return(0);
}


void play_multi(int argc,char *argv[],int argsleft)
{
int f,entnum;
struct stat buf;

/* construct a gsm directory consisting entirely of the files, but tagged */
for(f=argc-argsleft,entnum=1;f<argc;f++)
  if(stat(argv[f],&buf)!=-1 && !S_ISDIR(buf.st_mode))
    {
    entnum++;
    /* curses not started up at this point, so vanilla error exit is ok */
    if(!gsmdir_resize_if_needed(entnum))
      fprintf(stderr,"zgsmplay: out of memory\n"),exit(1);
    strcpy(gsmdir[entnum].name,argv[f]);
    gsmdir[entnum].marked=1;
    gsmdir[entnum].isdir=0;
    }

gsmdirsiz=entnum;
load_tagged_files();	/* play them */
}


void filesel()
{
int quit,key,curent;
int oldent,startfrom,oldstart,f,markchange;

quit=0; curent=1; oldent=1;
startfrom=1;

fselheadfoot();
readgsmdir();
showgsmdir(startfrom,0);
showbar(curent,1,startfrom);

while(!quit)
  {
  oldent=curent;
  markchange=0;
  usleep(20000);
  key=readnbkey();
  
  switch(key)
    {
    case 18: case 12:   /* ^R and ^L */
      readgsmdir();
      if(curent>gsmdirsiz) curent=gsmdirsiz;
      oldent=curent;
      redrawall(curent,startfrom);
      break;
    case RK_HOME: case 'A'-0x40: /* home or ^a */
      curent=1;
      break;
    case RK_END:  case 'E'-0x40: /* end  or ^e */
      curent=gsmdirsiz;
      break;
    case RK_PAGE_UP: case 'U'-0x40: /* pgup or ^u */
      curent-=YSIZ*(XSIZ-1);
      if(curent<1) curent=1;
      break;
    case RK_PAGE_DOWN: case 'V'-0x40: /* pgdn or ^v */
      curent+=YSIZ*(XSIZ-1);
      if(curent>gsmdirsiz) curent=gsmdirsiz;
      break;
    case 'D': case RK_DELETE:   /* shift-D or the 'del' key */
      if(gsmdir[curent].isdir) break;
      if(delete_file(gsmdir[curent].name))
        {
        readgsmdir();
        if(curent>gsmdirsiz) curent=gsmdirsiz;
        oldent=curent;
        }
      redrawall(curent,startfrom);
      break;
    case 't':		/* tag current file and move on */
    case '+': case '=':
      if(gsmdir[curent].isdir==0)
        {
        gsmdir[curent].marked=1;
        markchange=1;
        }
      if(curent<gsmdirsiz) curent++;
      break;
    case 'n': 		/* untag current file and move on */
    case 'u': case '-':
      if(gsmdir[curent].isdir==0)
        {
        gsmdir[curent].marked=0;
        markchange=1;
        }
      if(curent<gsmdirsiz) curent++;
      break;
    case ' ':   /* toggle tag/untag flag of current file and move on */
      if(gsmdir[curent].isdir==0)
        {
        gsmdir[curent].marked=1-gsmdir[curent].marked;
        markchange=1;
        }
      if(curent<gsmdirsiz) curent++;
      break;
    case 'T':	/* tag all files */
      for(f=1;f<=gsmdirsiz;f++) if(gsmdir[f].isdir==0) gsmdir[f].marked=1;
      redrawall(curent,startfrom);
      break;
    case 'N':	/* untag all files */
    case 'U':
      for(f=1;f<=gsmdirsiz;f++) if(gsmdir[f].isdir==0) gsmdir[f].marked=0;
      redrawall(curent,startfrom);
      break;
    case 'q': case 'k': case RK_CURSOR_UP:
      if(curent>1) curent--; break;
    case 'a': case 'j': case RK_CURSOR_DOWN:
      if(curent<gsmdirsiz) curent++; break;
    case 'o': case 'h': case RK_CURSOR_LEFT:
      curent-=YSIZ;
      if(curent<1) curent=1;
      break;
    case 'p': case 'l': case RK_CURSOR_RIGHT:
      curent+=YSIZ;
      if(curent>gsmdirsiz) curent=gsmdirsiz;
      break;
    case RK_ENTER: case 13:
      /* uhhhhh my head hurts */
      sgres=0;
      do
        {
        if(permissiondenied(gsmdir[curent].name))
          {
          redrawall(curent,startfrom);
          pic_incr=0;	/* stop here *even if ^p or ^n used* */
          break;
          }
          
        if(gsmdir[curent].isdir)
          {
          showbar(curent,0,startfrom);
          showgsmdir(startfrom,1);
          /* this may take a little while, so put cursor in sane place */
          move(fwinypos(1),fwinxpos(1)+1);
          refresh();
          chdir(gsmdir[curent].name);
          readgsmdir();
          showgsmdir(1,0); curent=oldent=startfrom=1;
          showbar(curent,1,startfrom);
          break;
          }
          
        tagview_mode=0;
        if(load_single_file(curent)==0) pic_incr=0;

	if(sgres==3) gsmdir[curent].marked=1;
        curent+=pic_incr;
        if(curent>gsmdirsiz) { curent--; pic_incr=0; }
        if(curent<1 || gsmdir[curent].isdir) { curent++; pic_incr=0; }
        oldstart=startfrom;
        while(curent<startfrom)
          startfrom-=YSIZ;
        while(fwinxpos(curent-startfrom+1)+BARWIDTH>COLS)
          startfrom+=YSIZ;
        if(startfrom<1) startfrom=1;
        if(sgres<2 || pic_incr==0)
          {
          oldent=-1;
          redrawall(curent,startfrom);
          }
        }
      while(pic_incr);
      break;
      
    case '\t':		/* tab */
      load_tagged_files();
      redrawall(curent,startfrom);
      break;
    case 'x': case RK_ESC:
      quit=1;
    }

  /* if we received SIGWINCH, redraw `frame' before updating stuff.
   * There's a race condition here of course, but I doubt it's
   * likely to be a problem in practice.
   */
  if(window_resized)
    {
    window_resized=0;
    fix_scrn_size();
    redrawall(curent,startfrom);
    oldent=curent;
    }
  
  oldstart=startfrom;
  while(curent<startfrom)
    startfrom-=YSIZ;
  while(fwinxpos(curent-startfrom+1)+BARWIDTH>COLS)
    startfrom+=YSIZ;
  if(startfrom<1) startfrom=1;
  if(startfrom!=oldstart)
    {
    showbar(oldent,0,oldstart);
    showgsmdir(oldstart,1);
    showgsmdir(startfrom,0);
    showbar(curent,1,startfrom);
    }
  else  
    if(oldent!=-1 && (curent!=oldent || markchange))
      {
      showbar(oldent,0,startfrom);
      showbar(curent,1,startfrom);
      }
  }
}
